/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static java.lang.Math.max;
import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.ui.editor.fx.visual.CoordinateCorrections.RECTANGLE_INSETS;
import static org.fortiss.tooling.base.ui.utils.ContextMenuUtils.createElementCompositionContext;
import static org.fortiss.tooling.base.ui.utils.ContextMenuUtils.createPrototypeMenu;
import static org.fortiss.tooling.base.ui.utils.FXDNDUtils.canCompose;
import static org.fortiss.tooling.base.ui.utils.FXDNDUtils.compose;

import java.util.List;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.ElementPackage;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IClickController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.ClickControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.rectangular.RectangularContentAnchorageMoveController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.rectangular.RectangularResizableContentControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

import javafx.geometry.Dimension2D;
import javafx.scene.Node;
import javafx.scene.control.MenuItem;
import javafx.scene.input.Dragboard;
import javafx.scene.input.MouseEvent;

/**
 * {@link RectangularResizableContentControllerBase} with a listener mechanism for the layout of the
 * connected {@link EObject model element}.
 * 
 * @author hoelzl
 */
public abstract class EObjectRectangularResizableContentControllerBase<T extends INamedCommentedElement & ILayoutedModelElement>
		extends RectangularResizableContentControllerBase {

	/** {@link IModelChangeProvider} for this controller */
	private final NamedCommentedLayoutModelChangeProvider<T> layoutModelChangeProvider;
	/** The click controller handling double-click to open editor. */
	private final IClickController openEditorDoubleClickController = new ClickControllerBase() {
		@Override
		public Change singleClick(MouseEvent event, Node node, DiagramCoordinate diagramLocation) {
			return EObjectRectangularResizableContentControllerBase.super.getClickController(node,
					diagramLocation).singleClick(event, node, diagramLocation);
		}

		@Override
		public Change secondaryClick(MouseEvent event, Node node,
				DiagramCoordinate diagramLocation) {
			return EObjectRectangularResizableContentControllerBase.super.getClickController(node,
					diagramLocation).secondaryClick(event, node, diagramLocation);
		}

		@Override
		public Change doubleClick(MouseEvent event, Node node, DiagramCoordinate diagramLocation) {
			EObject eo = getModelElement();
			IModelElementHandler<EObject> handler =
					IModelElementHandlerService.getInstance().getModelElementHandler(eo);
			if(handler != null) {
				eo = handler.handleOpenModelElementRequest(eo);
			}
			IModelEditorBindingService.getInstance().openInEditor(eo);
			return null; // no changes to model
		}
	};

	/** Constructor. */
	public EObjectRectangularResizableContentControllerBase(IContentMVCBundle mvcb,
			Class<T> modelType) {
		super(mvcb);

		Object model = requireNonNull(mvcb.getModel(), "The given model is null!");
		if(!modelType.isAssignableFrom(model.getClass())) {
			throw new IllegalArgumentException("Expected model of type " +
					modelType.getSimpleName() + ", but was " + model.getClass().getSimpleName());
		}

		@SuppressWarnings("unchecked") T lme = (T)mvcb.getModel();
		this.layoutModelChangeProvider = new NamedCommentedLayoutModelChangeProvider<T>(lme) {
			/** {@inheritDoc} */
			@Override
			protected boolean acceptNotification(Notification notification) {
				if(notification.getNotifier() != getModelElement()) {
					return false;
				}
				return isAnchorageFeature(notification.getFeature());
			}
		};
	}

	/** Returns the correctly casted model element. */
	@SuppressWarnings("unchecked")
	protected T getModelElement() {
		// wild cast works: see constructor check with exception
		return (T)getModel();
	}

	/** {@inheritDoc} */
	@Override
	public IModelChangeProvider getModelChangeProvider() {
		return layoutModelChangeProvider;
	}

	/** {@inheritDoc} */
	@Override
	protected IDragController createAnchorageMoveController(IContentAnchorageMVCBundle anchorage) {
		// we need an overridden port move controller due to the frameworks default snapping
		// behavior, which snaps to anchorage size while AF3 ports snap to half their size
		return new RectangularContentAnchorageMoveController(this, anchorage) {
			/** {@inheritDoc} */
			@Override
			protected Dimension2D overrideBorderSnap(Dimension2D anchorageSize) {
				// this override is required because ports can be moved by half their size
				// while the default implementation would only allow for full size of the port
				return new Dimension2D(anchorageSize.getWidth() / 2, anchorageSize.getHeight() / 2);
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	protected Dimension2D getMinimumSize() {
		return new Dimension2D(DEFAULT_SHAPE_MINIMUM_WIDTH - 2 * RECTANGLE_INSETS.getWidth(),
				DEFAULT_SHAPE_MINIMUM_HEIGHT - 2 * RECTANGLE_INSETS.getHeight());
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		// outer-most grid areas of the diagram are inaccessible
		DiagramViewerFeatures features = getViewer().getFeatures();
		double x = max(features.getHorizontalSpacing(), diagramLocation.getX());
		double y = max(features.getVerticalSpacing(), diagramLocation.getY());

		// wild cast works: see constructor exception
		EObject modelParent = (EObject)getModel();
		IElementCompositionContext edc = createElementCompositionContext(modelParent, x, y, false,
				getViewer().getFeatures().getCurrentZoomFactor());
		return createPrototypeMenu(modelParent, edc);
	}

	/** {@inheritDoc} */
	@Override
	public void delete() {
		IElementCompositorService.getInstance().decompose(getModelElement());
	}

	/** Checks whether the given feature corresponds to an anchorage model element. */
	protected boolean isAnchorageFeature(Object feature) {
		return feature == ElementPackage.Literals.IHIERARCHIC_ELEMENT__CONNECTORS;
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		return openEditorDoubleClickController;
	}

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDragOver(Dragboard db, DiagramCoordinate diagramLocation) {
		double zoom = getViewer().getFeatures().getCurrentZoomFactor();
		return canCompose(db, diagramLocation, getModelElement(), false, zoom);
	}

	/** {@inheritDoc} */
	@Override
	public boolean handleExternalDNDDrop(Dragboard db, DiagramCoordinate diagramLocation) {
		double zoom = getViewer().getFeatures().getCurrentZoomFactor();
		return compose(db, diagramLocation, getModelElement(), false, zoom);
	}
}
