/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.model;

import static java.util.Collections.emptyList;
import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.common.util.LambdaUtils.filterList;

import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelFactory;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IHierarchicElement;

/**
 * {@link IModelFactory} for {@link IHierarchicElement}s of editors.
 * 
 * @author diewald
 */
public abstract class HierarchicElementModelFactoryBase<T extends IHierarchicElement>
		implements IModelFactory {
	/** The currently edited element. */
	private final T rootElement;

	/** Constructor. */
	public HierarchicElementModelFactoryBase(T root) {
		this.rootElement = requireNonNull(root);
	}

	/** {@inheritDoc} */
	@Override
	public T getRootModel() {
		return rootElement;
	}

	/** {@inheritDoc} */
	@Override
	public List<?> getDiagramAnchorageModels() {
		return rootElement.getConnectors();
	}

	/** {@inheritDoc} */
	@Override
	public List<?> getContentAnchorageModels(Object parent) {
		if(parent instanceof IHierarchicElement) {
			return ((IHierarchicElement)parent).getConnectors();
		}
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public List<?> getLinkModels() {
		return filterList(rootElement.getConnections(),
				c -> filterPartiallyInitializedConnections(c));
	}

	/** Checks whether the given connection is only partially initialized. */
	private boolean filterPartiallyInitializedConnections(IConnection c) {
		return c.getSource() != null && c.getTarget() != null && c.eContainer() != null;
	}

	/** {@inheritDoc} */
	@Override
	public Object getLinkStart(Object link) {
		if(link instanceof IConnection) {
			return ((IConnection)link).getSource();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Object getLinkEnd(Object link) {
		if(link instanceof IConnection) {
			return ((IConnection)link).getTarget();
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Object getParent(Object element) {
		// Not used at the moment, so return null.
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void update() {
		// Not used at the moment.
	}
}
