/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.migration;

import static java.util.Collections.emptyMap;
import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.fortiss.tooling.common.util.LambdaUtils.isAssignableFromAny;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.IMigrationProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;

/**
 * {@link IMigrationProvider} that ensures that the annotation of the given {@code annotationType}
 * is only kept for the {@link IModelElement} specializations of the given. From all other
 * {@link IModelElement} specializations, the annotation instance is removed.
 * 
 * @author barner
 */
public abstract class RemoveOutdatedAnnotationInstanceMigrationProvider<T extends IAnnotatedSpecification>
		implements IMigrationProvider {
	/** Annotation type for which outdated instances should be removed. */
	private Class<T> annotationType;

	/** List of {@link IModelElement} types for which annotation should be kept. */
	private List<Class<?>> modelElementTypes;

	/** Constructor. */
	protected RemoveOutdatedAnnotationInstanceMigrationProvider(Class<T> annotationType,
			List<Class<?>> modelElementTypes) {
		this.annotationType = annotationType;
		this.modelElementTypes = modelElementTypes;
	}

	/** {@inheritDoc} */
	@Override
	public boolean needMigration(ITopLevelElement modelElement,
			Map<EObject, AnyType> unknownFeatures) {

		EObject rootModelElement = modelElement.getRootModelElement();
		for(T annotation : getChildrenWithType(rootModelElement, annotationType)) {
			if(!isAssignableFromAny(modelElementTypes, annotation.getSpecificationOf())) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Map<EObject, AnyType> migrate(ITopLevelElement modelElement,
			Map<EObject, AnyType> unknownFeatures) {
		EObject rootModelElement = modelElement.getRootModelElement();
		for(T annotation : getChildrenWithType(rootModelElement, annotationType)) {
			if(!isAssignableFromAny(modelElementTypes, annotation.getSpecificationOf())) {
				delete(annotation);
			}
		}
		return emptyMap();
	}
}
