/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx;

import static javafx.scene.layout.AnchorPane.setBottomAnchor;
import static javafx.scene.layout.AnchorPane.setLeftAnchor;
import static javafx.scene.layout.AnchorPane.setRightAnchor;
import static javafx.scene.layout.AnchorPane.setTopAnchor;

import java.net.URL;

import org.apache.commons.lang3.SystemUtils;
import org.eclipse.fx.ui.workbench3.FXViewPart;
import org.fortiss.tooling.common.ui.javafx.layout.ICompositeFXController;

import javafx.application.Platform;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.Pane;

/**
 * Base class to create eclipse (SWT-based) views using JavaFX GUI elements.
 * 
 * @author munaro
 */
public abstract class AF3FXViewPart extends FXViewPart {

	/** For later use, if we need to define different styles. */
	private String cssLocation;

	/** Flag that indicates if a custom style sheet has been applied. */
	private boolean isCustomStyleSheetApplied;

	/** References the root {@link Pane} to which GUI elements must be added to be displayed. */
	private Pane root;

	/** Specifies the view part's layout and logic. */
	protected final ICompositeFXController<? extends Node, ? extends Node> controller;

	/**
	 * Constructor. Allows to pass the view's FXML definition file.
	 * 
	 * @param controller
	 *            {@link ICompositeFXController} specifying the layout of the view part as well as
	 *            its logic.
	 * @param cssLocation
	 *            style sheet file path for the appearance settings relative to the source
	 *            (resource) folders. Pass {@code null} if the OS-native look shall be used.
	 * @throws Exception
	 *             if the given FXML file path is not given or if it cannot be found in any of the
	 *             source (resource) folders.
	 */
	public AF3FXViewPart(ICompositeFXController<? extends Node, ? extends Node> controller,
			String cssLocation) throws Exception {
		Platform.setImplicitExit(false);
		this.cssLocation = cssLocation;
		this.controller = controller;
	}

	/** {@inheritDoc} */
	@Override
	protected Scene createFxScene() {
		controller.loadLayout();
		Node layout = controller.getLayout();

		setTopAnchor(layout, 0.0);
		setRightAnchor(layout, 0.0);
		setBottomAnchor(layout, 0.0);
		setLeftAnchor(layout, 0.0);

		root = new AnchorPane(layout);
		if(cssLocation != null) {
			isCustomStyleSheetApplied = applyStyleSheet(root, getClass(), cssLocation);
		}
		if(!isCustomStyleSheetApplied) {
			setNativeLook(root);
		}
		Scene scene = new Scene(root);
		return scene;
	}

	/** {@inheritDoc} */
	@Override
	protected void setFxFocus() {
		// Not needed.
	}

	/**
	 * Sets the native look of the target platform to not break the UI experience. Most of the UI
	 * code is based on SWT which uses a platform native look. Currently, Windows 8&10, Windows 7,
	 * and MacOS X native looks are supported.
	 */
	private void setNativeLook(Pane pane) {
		if(SystemUtils.IS_OS_WINDOWS_10 || SystemUtils.IS_OS_WINDOWS_8) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/jfxtras/JMetroLightTheme.css");
		} else if(SystemUtils.IS_OS_WINDOWS) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/aerofx/win7.css");
		} else if(SystemUtils.IS_OS_MAC_OSX) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/aquafx/aquafx.css");
		} else if(SystemUtils.IS_OS_LINUX) {
			// At the time of writing this method, no CSS is known that mimics GTK or QT. Thus, use
			// the default modena look.
		}
	}

	/**
	 * Applies the style sheet from the given location to the given {@link Pane}. The class loader
	 * of the plugin containing the style sheet must be given such that this resource can be loaded.
	 * For instance the class loader of the common.ui plugin must be given for the native looks.
	 * 
	 * @param pane
	 *            {@link Pane} to which the style sheet shall be applied.
	 * @param viewerClass
	 *            Class containing the style sheet resource.
	 * @param styleSheetLocation
	 *            relative location of the style sheet to its containing resource folder (aka source
	 *            folder).
	 * @return whether the style sheet could be applied.
	 */
	private boolean applyStyleSheet(Pane pane, Class<?> viewerClass, String styleSheetLocation) {
		if(viewerClass == null || styleSheetLocation == null) {
			return false;
		}
		URL cssURL = viewerClass.getResource(styleSheetLocation);
		if(cssURL != null) {
			pane.getStylesheets().add(cssURL.toExternalForm());
			return true;
		}
		return false;
	}
}
