/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.rectangular;

import static java.lang.Math.hypot;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.AnchorageContentControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.ContentAnchorageMoveControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;

/** Base class for controllers with content visuals represented by diamonds. */
public abstract class DiamondResizableContentControllerBase extends AnchorageContentControllerBase {
	/** Relative X coordinates to be checked for anchorage locations. */
	private static final double[] XCOORDS = {0.5, 0.5, 0, 1};
	/** Relative Y coordinates to be checked for anchorage locations. */
	private static final double[] YCOORDS = {0, 1, 0.5, 0.5};

	/** Constructor. */
	public DiamondResizableContentControllerBase(IContentMVCBundle mvcb) {
		super(mvcb);
	}

	/** Finds the closest corner of the diamond shape relative to the given indicator position. */
	private Side getClosestBorderLocation(double indicatorX, double indicatorY,
			Rectangle2D bounds) {
		Side resultSide = Side.values()[0];
		double hypo = Double.MAX_VALUE;
		double boundsX = bounds.getMinX();
		double boundsY = bounds.getMinY();
		for(int cur = 0; cur < XCOORDS.length; cur++) {
			double anchorageX = boundsX + getXCoord(cur, bounds);
			double anchorageY = boundsY + getYCoord(cur, bounds);
			double curHypo = hypot(indicatorX - anchorageX, indicatorY - anchorageY);
			if(curHypo < hypo) {
				hypo = curHypo;
				resultSide = Side.values()[cur];
			}
		}
		return resultSide;
	}

	/** Computes X coordinate. */
	private double getXCoord(int cur, Rectangle2D b) {
		return b.getWidth() * XCOORDS[cur];
	}

	/** Computes Y coordinate. */
	private double getYCoord(int cur, Rectangle2D b) {
		return b.getHeight() * YCOORDS[cur];
	}

	/** {@inheritDoc} */
	@Override
	protected IDragController createAnchorageMoveController(IContentAnchorageMVCBundle anchorage) {
		return new ContentAnchorageMoveControllerBase(anchorage) {
			@Override
			protected Point2D getAnchorageFeedbackLocation(DiagramCoordinate delta) {
				Rectangle2D cBounds = getVisual().getModelBounds();
				Rectangle2D aBounds = anchorage.getVisual().getModelBounds();
				double indicatorX = cBounds.getMinX() + aBounds.getMinX() + delta.getX();
				double indicatorY = cBounds.getMinY() + aBounds.getMinY() + delta.getY();
				Side closest = getClosestBorderLocation(indicatorX, indicatorY, cBounds);
				int idx = closest.ordinal();
				double x = cBounds.getWidth() * XCOORDS[idx] - aBounds.getWidth() / 2;
				double y = cBounds.getHeight() * YCOORDS[idx] - aBounds.getHeight() / 2;
				return new Point2D(x, y);
			}

			@Override
			protected void moveAnchorage(DiagramCoordinate delta) {
				Rectangle2D cBounds = getVisual().getModelBounds();
				Rectangle2D aBounds = anchorage.getVisual().getModelBounds();
				double indicatorX = cBounds.getMinX() + aBounds.getMinX() + delta.getX();
				double indicatorY = cBounds.getMinY() + aBounds.getMinY() + delta.getY();
				moveAnchorageToSideOffset(anchorage,
						getClosestBorderLocation(indicatorX, indicatorY, cBounds), 0);
			}
		};
	}

	/** Sets the side and offset of the given anchorage when move operation has completed. */
	protected abstract void moveAnchorageToSideOffset(IContentAnchorageMVCBundle anchorage,
			Side side, double offset);
}
