/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_INCOMING_LINK_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_OUTGOING_LINK_TAG;

import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.MVCBundleTag;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.widgets.ExpandCollapseWidget;

import javafx.geometry.Insets;
import javafx.geometry.Rectangle2D;
import javafx.geometry.VPos;
import javafx.scene.Node;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.shape.Shape;
import javafx.scene.text.Text;
import javafx.scene.text.TextAlignment;

/** Base implementation for {@link IContentVisual}. */
public abstract class ContentVisualBase extends VisualBase implements IContentVisual {
	/** The text of this visual. */
	protected final Text text = new Text();
	/** The icon of this visual. */
	protected final ImageView icon = new ImageView();
	/** The expanded / collapsed indicator widget. */
	protected final ExpandCollapseWidget expandCollapseWidget =
			new ExpandCollapseWidget(0, 0, 20, 20, 3);

	/** Constructor. */
	public ContentVisualBase(IContentMVCBundle mvcb, Shape visual, Shape hit) {
		super(mvcb, visual, hit);
	}

	/** Determines whether the icon is enabled or not. */
	protected boolean enableIcon() {
		return true;
	}

	/** Determines whether the text of this visual is enabled or not. */
	protected boolean enableText() {
		return true;
	}

	/** Determines whether the expand/collapse widget of this visual is enabled or not. */
	protected boolean enableExpandCollapseWidget() {
		return false;
	}

	/** Returns the name of the content. */
	protected String getName() {
		return "";
	}

	/** Returns the anchor location relative to the parent bounds. */
	protected DiagramCoordinate getTextAnchorLocation() {
		Insets i = getTextInsets();
		double croppedHeight = getCurrentBounds().getHeight() - i.getTop() - i.getBottom();
		return new DiagramCoordinate(0, croppedHeight / 2);
	}

	/** Returns the insets of the text label. */
	protected Insets getTextInsets() {
		// TODO (SB): Magic constant
		return new Insets(10, 10, 10, 10);
	}

	/** Returns the vertical alignment of the text label. */
	protected VPos getVerticalTextAlignment() {
		return VPos.CENTER;
	}

	/** Returns the horizontal alignment of the text label. */
	protected TextAlignment getHorizontalTextAlignment() {
		return TextAlignment.LEFT;
	}

	/**
	 * Returns the icon image for the content. Sub-classes should use an image cache instead of
	 * creating a new image every time this method is called.
	 */
	protected Image getIcon() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public final void applySelectedFocusedEffect(DiagramLayers layers) {
		super.applySelectedFocusedEffect(layers);

		if(getViewer().getFeatures().isLinkHighlightingEnabled()) {
			IContentMVCBundle bundle = getMVCBundle();
			List<IContentAnchorageMVCBundle> anchoragesBundles = bundle.getAnchorages();
			for(IContentAnchorageMVCBundle aBundle : anchoragesBundles) {
				List<ILinkMVCBundle> incomingLinksBundles = aBundle.getIncomingLinks();
				for(ILinkMVCBundle ilBundle : incomingLinksBundles) {
					addTag(ilBundle, HIGHLIGHT_INCOMING_LINK_TAG, layers);
					addTag(ilBundle.getStartAnchorage(), HIGHLIGHT_INCOMING_LINK_TAG, layers);
					addTag(ilBundle.getEndAnchorage(), HIGHLIGHT_INCOMING_LINK_TAG, layers);
				}
				List<ILinkMVCBundle> outgoingLinksBundles = aBundle.getOutgoingLinks();
				for(ILinkMVCBundle olBundle : outgoingLinksBundles) {
					addTag(olBundle, HIGHLIGHT_OUTGOING_LINK_TAG, layers);
					addTag(olBundle.getEndAnchorage(), HIGHLIGHT_OUTGOING_LINK_TAG, layers);
					addTag(olBundle.getStartAnchorage(), HIGHLIGHT_OUTGOING_LINK_TAG, layers);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void applyNotSelectedEffect(DiagramLayers layers) {
		super.applyNotSelectedEffect(layers);

		IContentMVCBundle bundle = getMVCBundle();
		List<IContentAnchorageMVCBundle> anchoragesBundles = bundle.getAnchorages();
		for(IContentAnchorageMVCBundle aBundle : anchoragesBundles) {
			List<ILinkMVCBundle> incomingLinksBundles = aBundle.getIncomingLinks();
			for(ILinkMVCBundle ilBundle : incomingLinksBundles) {
				removeTag(ilBundle, HIGHLIGHT_INCOMING_LINK_TAG, layers);
				removeTag(ilBundle.getStartAnchorage(), HIGHLIGHT_INCOMING_LINK_TAG, layers);
				removeTag(ilBundle.getEndAnchorage(), HIGHLIGHT_INCOMING_LINK_TAG, layers);
			}
			List<ILinkMVCBundle> outgoingLinksBundles = aBundle.getOutgoingLinks();
			for(ILinkMVCBundle olBundle : outgoingLinksBundles) {
				removeTag(olBundle, HIGHLIGHT_OUTGOING_LINK_TAG, layers);
				removeTag(olBundle.getEndAnchorage(), HIGHLIGHT_OUTGOING_LINK_TAG, layers);
				removeTag(olBundle.getStartAnchorage(), HIGHLIGHT_OUTGOING_LINK_TAG, layers);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getCurrentBounds() {
		FeedbackChange fb = getFeedbackChange();
		Rectangle2D mb = getModelBounds();
		return fb.applyTo(mb);
	}

	/** Adds a tag for the given bundle and updates the visual. */
	private void addTag(IMVCBundle bundle, MVCBundleTag tag, DiagramLayers layers) {
		bundle.addTag(tag);
		bundle.getVisual().updateNodes(layers);
	}

	/** Removes a tag from the given bundle and updates the visual. */
	private void removeTag(IMVCBundle bundle, MVCBundleTag tag, DiagramLayers layers) {
		bundle.removeTag(tag);
		bundle.getVisual().updateNodes(layers);
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Sub-classes may override, but must call super implementation to update children properly.
	 */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		IContentMVCBundle bundle = getMVCBundle();
		Rectangle2D bounds = getCurrentBounds();
		double leftX = bounds.getMinX();
		double rightX = bounds.getMaxX();
		double upperY = bounds.getMinY();

		Insets i = getTextInsets();
		DiagramCoordinate anchorLocation = getTextAnchorLocation();
		double anchorX = leftX + anchorLocation.getX() + i.getLeft();
		double anchorY = upperY + anchorLocation.getY() + i.getTop();

		if(enableIcon()) {
			if(icon.getParent() == null) {
				layers.getTextLayer().add(icon, bundle);
			}
			Image img = getIcon();
			icon.setImage(img);
			if(img != null) {
				icon.setX(anchorX);
				anchorX += img.getWidth() + i.getLeft();
				icon.setY(anchorY - img.getHeight() / 2);
			}
		} else {
			layers.getTextLayer().remove(icon);
		}

		if(enableText()) {
			if(text.getParent() == null) {
				layers.getTextLayer().add(text, bundle);
			}
			text.setX(anchorX);
			text.setY(anchorY);
			text.setText(getName());
			text.setWrappingWidth(rightX - anchorX - i.getRight());
			text.setTextAlignment(getHorizontalTextAlignment());
			text.setTextOrigin(getVerticalTextAlignment());
		} else {
			layers.getTextLayer().remove(text);
		}

		if(enableExpandCollapseWidget()) {
			if(expandCollapseWidget.getParent() == null) {
				layers.getContentInteractionLayer().add(expandCollapseWidget, bundle);
			}
			Rectangle2D ecWidgetLocation = getExpandCollapseWidgetRectangle();
			double inset = getExpandCollapseWidgetInset();
			double x = leftX + ecWidgetLocation.getMinX();
			double y = upperY + ecWidgetLocation.getMinY();
			expandCollapseWidget.update(x, y, ecWidgetLocation.getWidth(),
					ecWidgetLocation.getHeight(), inset);
			expandCollapseWidget.setState(getExpandCollapseWidgetState());
		} else {
			layers.getContentInteractionLayer().remove(expandCollapseWidget);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		super.removeAllVisuals(layers);

		layers.getContentInteractionLayer().remove(expandCollapseWidget);
		layers.getTextLayer().remove(text);
		layers.getTextLayer().remove(icon);
	}

	/** Returns the current state of the expand/collapse widget. */
	protected boolean getExpandCollapseWidgetState() {
		return true;
	}

	/** Returns the inset for the plus/minus sign of the expand/collapse widget. */
	protected double getExpandCollapseWidgetInset() {
		// TODO (SB): Magic constant
		return 3;
	}

	/** Returns the location for the expand/collapse widget. */
	protected Rectangle2D getExpandCollapseWidgetRectangle() {
		Rectangle2D b = getCurrentBounds();
		// TODO (SB): Magic constant
		return new Rectangle2D(b.getWidth() - 20, 6, 14, 14);
	}

	/** {@inheritDoc} */
	@Override
	public IContentMVCBundle getMVCBundle() {
		// wild cast works: see constructor
		return (IContentMVCBundle)super.getMVCBundle();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isExpandCollapseWidget(Node widget) {
		return expandCollapseWidget.isHit(widget);
	}

	/** {@inheritDoc} */
	@Override
	public boolean updateAttachedAnchorageVisualsOnHover() {
		return false;
	}
}
