/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic;

import static java.lang.Math.cos;
import static java.lang.Math.hypot;
import static java.lang.Math.max;
import static java.lang.Math.sin;
import static java.lang.Math.toRadians;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.layout.IAngleLayout;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base.ContentVisualBase;

import javafx.geometry.Bounds;
import javafx.geometry.Dimension2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;

/** Base class for {@link ContentVisualBase content visuals} depicted by rectangles. */
public abstract class CircularContentVisualBase extends ContentVisualBase {
	/** Constructor. */
	public CircularContentVisualBase(IContentMVCBundle mvcb) {
		super(mvcb, new Circle(), new Circle());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		double leftX = bounds.getMinX();
		double upperY = bounds.getMinY();
		double width = bounds.getWidth();
		double height = bounds.getHeight();
		double w2 = width / 2.0;
		double cx = leftX + w2;
		double h2 = height / 2.0;
		double cy = upperY + h2;
		double radius = max(w2, h2);
		if(enableVisual()) {
			Circle circle = getVisualCircleShape();
			circle.setCenterX(cx);
			circle.setCenterY(cy);
			circle.setRadius(radius);
			circle.setOpacity(getOpacity());
			circle.setStroke(getBorderColor());
			circle.setStrokeWidth(getBorderWidth());
			circle.setStrokeType(getBorderType());
			circle.setFill(getFillColor());
		}

		if(enableInteraction()) {
			Circle ha = getHitAreaCircleShape();
			double lnk = getHitAreaStartLinkSize();
			double resize = getHitAreaResizeSize();
			radius = lnk + resize + radius;
			ha.setCenterX(cx);
			ha.setCenterY(cy);
			ha.setRadius(radius);

			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getAnchorageLocation(IContentAnchorageVisual visual) {
		Rectangle2D pb = getCurrentBounds();
		double w2 = pb.getWidth() / 2.0;
		double h2 = pb.getHeight() / 2.0;
		double radius = max(w2, h2);
		double angleDeg = visual.getLayout(IAngleLayout.class).getAngleInDegree();
		double angleRad = toRadians(angleDeg);
		double nx = cos(angleRad);
		double ny = sin(angleRad);
		double x = pb.getMinX() + w2 + radius * nx;
		double y = pb.getMinY() + h2 + radius * ny;
		return new DiagramCoordinate(x, y);
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		return getLocationOnCircle(indication);
	}

	/** Returns the location on the circle closest to the given indication coordinate. */
	private DiagramCoordinate getLocationOnCircle(DiagramCoordinate indication) {
		Circle circ = getVisualCircleShape();
		double radius = getRadius(circ);
		double cx = circ.getCenterX();
		double dx = indication.getX() - cx;
		double cy = circ.getCenterY();
		double dy = indication.getY() - cy;
		double len = hypot(dx, dy);
		double nx = radius * dx / len;
		double ny = radius * dy / len;
		return new DiagramCoordinate(cx + nx, cy + ny);
	}

	/** Returns the dimensions of the corner arcs. */
	protected Dimension2D getCornerArcDimensions() {
		return new Dimension2D(15, 15);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		Circle circ = getHitAreaCircleShape();
		if(node == circ) {
			double dx = diagramLocation.getX() - circ.getCenterX();
			double dy = diagramLocation.getY() - circ.getCenterY();
			double radius = getRadius(circ);
			double hypot = hypot(dx, dy);
			if(hypot < radius - getHitAreaResizeSize() - getHitAreaStartLinkSize()) {
				return EDragGesture.MOVE;
			}
			if(hypot < radius - getHitAreaStartLinkSize()) {
				return EDragGesture.RESIZE_VH;
			}
			if(hypot < radius) {
				return EDragGesture.NEW_LINK;
			}
		}
		if(node == getVisualShape() || node == text) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/** Returns the radius of the given circle computed from node bounds. */
	private double getRadius(Circle circ) {
		Bounds bounds = circ.getBoundsInLocal();
		double w2 = bounds.getWidth() / 2.0;
		double h2 = bounds.getHeight() / 2.0;
		double radius = max(w2, h2);
		return radius;
	}

	/** Returns the visual circle shape. */
	protected final Circle getVisualCircleShape() {
		// wild cast works: see constructor
		return (Circle)getVisualShape();
	}

	/** Returns the hit area circle shape. */
	protected final Circle getHitAreaCircleShape() {
		// wild cast works: see constructor
		return (Circle)getHitAreaShape();
	}

	/**
	 * Returns the extension of the hit area in pixel used to trigger link creation. This number
	 * specifies how many pixels beyond the visible border are used to trigger link creation by a
	 * mouse drag gesture.
	 */
	protected double getHitAreaStartLinkSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the size of the hit area in pixel used to resize the content visual. */
	protected double getHitAreaResizeSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}
}
