/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic;

import static javafx.scene.paint.Color.BLACK;
import static javafx.scene.paint.Color.GRAY;
import static javafx.scene.paint.Color.RED;
import static javafx.scene.paint.Color.TRANSPARENT;
import static javafx.scene.paint.Color.WHITE;
import static javafx.scene.shape.StrokeLineCap.BUTT;
import static javafx.scene.shape.StrokeLineJoin.ROUND;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers.ILayer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.widgets.LinkArrowWidget;

import javafx.geometry.Insets;
import javafx.geometry.Point2D;
import javafx.scene.control.Label;
import javafx.scene.layout.Background;
import javafx.scene.layout.BackgroundFill;
import javafx.scene.layout.CornerRadii;
import javafx.scene.paint.Paint;
import javafx.scene.shape.CubicCurve;
import javafx.scene.shape.Line;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.TextAlignment;

/**
 * This class encapsulates {@link CubicCurve}s usually needed for every segment: a visible one, a
 * bigger and invisible one for mouse pointer hit tests.
 */
final class CurveSegment {
	/** The visible curve. */
	private final CubicCurve visibleCurve;
	/** The clickable curve. */
	private final CubicCurve clickableCurve;
	/** The selection feedback handle rectangle for the first control point. */
	private final Rectangle firstControlPointHandle;
	/** The selection feedback handle rectangle for the second control point. */
	private final Rectangle secondControlPointHandle;
	/** The selection feedback handle rectangle for the bend point. */
	private final Rectangle bendPointHandle;
	/** The visual helper line from the start to the first control point. */
	private final Line helperStart;
	/** The visual helper line from the second control point to the end point. */
	private final Line helperEnd;
	/** The arrow widget for curved links. */
	private final LinkArrowWidget arrowWidget;
	/** Stores the label. */
	private Label label;

	/** Constructor. */
	public CurveSegment(double sx, double sy, double c1x, double c1y, double c2x, double c2y,
			double ex, double ey, double selectionStrokeWidth, double bendPointFeedbackSize,
			double controlPointFeedbackSize, boolean showArrow, boolean useLineArrow,
			double arrowLength, String labelText) {
		visibleCurve = new CubicCurve(sx, sy, c1x, c1y, c2x, c2y, ex, ey);
		visibleCurve.setStroke(BLACK);
		visibleCurve.setFill(null);

		clickableCurve = new CubicCurve(sx, sy, c1x, c1y, c2x, c2y, ex, ey);
		clickableCurve.setStrokeWidth(selectionStrokeWidth);
		clickableCurve.setFill(null);
		clickableCurve.setStrokeLineJoin(ROUND);
		clickableCurve.setStrokeLineCap(BUTT);
		clickableCurve.setStroke(TRANSPARENT);

		double fs2 = bendPointFeedbackSize / 2;
		bendPointHandle =
				new Rectangle(ex - fs2, ey - fs2, bendPointFeedbackSize, bendPointFeedbackSize);
		bendPointHandle.setFill(TRANSPARENT);
		bendPointHandle.setStroke(RED);

		double cps2 = controlPointFeedbackSize / 2;
		firstControlPointHandle = new Rectangle(c1x - cps2, c1y - cps2, controlPointFeedbackSize,
				controlPointFeedbackSize);
		firstControlPointHandle.setFill(TRANSPARENT);
		firstControlPointHandle.setStroke(GRAY);

		secondControlPointHandle = new Rectangle(c2x - cps2, c2y - cps2, controlPointFeedbackSize,
				controlPointFeedbackSize);
		secondControlPointHandle.setFill(TRANSPARENT);
		secondControlPointHandle.setStroke(GRAY);

		helperStart = new Line(sx, sy, c1x, c1y);
		helperStart.setStroke(GRAY);

		helperEnd = new Line(c2x, c2y, ex, ey);
		helperEnd.setStroke(GRAY);

		if(showArrow) {
			arrowWidget = new LinkArrowWidget(useLineArrow, ex, ey, c2x, c2y, arrowLength);
		} else {
			arrowWidget = null;
		}
		if(labelText != null) {
			label = new Label(labelText);
			label.setTextAlignment(TextAlignment.CENTER);
		} else {
			label = null;
		}
	}

	/** Returns the visible curve. */
	public CubicCurve getVisibleCurve() {
		return visibleCurve;
	}

	/** Returns the clickable curve. */
	public CubicCurve getClickableCurve() {
		return clickableCurve;
	}

	/** Returns the first control point feedback handle. */
	public Rectangle getFirstControlPointHandle() {
		return firstControlPointHandle;
	}

	/** Returns the second control point feedback handle. */
	public Rectangle getSecondControlPointHandle() {
		return secondControlPointHandle;
	}

	/** Returns the bend-point feedback handle. */
	public Rectangle getBendPointHandle() {
		return bendPointHandle;
	}

	/** Returns the helper line of the start. */
	public Line getHelperStart() {
		return helperStart;
	}

	/** Returns the helper line of the end. */
	public Line getHelperEnd() {
		return helperEnd;
	}

	/** Sets the color of the visible lines. */
	public void setStrokeColor(Paint color) {
		visibleCurve.setStroke(color);
		if(arrowWidget != null) {
			arrowWidget.setColor(color);
		}
	}

	/** Sets the color of the invisible, clickable curves. */
	public void setClickableLinesColor(Paint color) {
		clickableCurve.setStroke(color);
	}

	/** Updates the segment coordinates. */
	public void update(Point2D sp, Point2D cp1, Point2D cp2, Point2D ep,
			double bendPointFeedbackSize, double controlPointFeedbackSize, double arrowLength,
			String labelText) {
		double sx = sp.getX();
		double sy = sp.getY();
		visibleCurve.setStartX(sx);
		visibleCurve.setStartY(sy);
		double c1x = cp1.getX();
		double c1y = cp1.getY();
		visibleCurve.setControlX1(c1x);
		visibleCurve.setControlY1(c1y);
		double c2x = cp2.getX();
		double c2y = cp2.getY();
		visibleCurve.setControlX2(c2x);
		visibleCurve.setControlY2(c2y);
		double ex = ep.getX();
		double ey = ep.getY();
		visibleCurve.setEndX(ex);
		visibleCurve.setEndY(ey);

		clickableCurve.setStartX(sx);
		clickableCurve.setStartY(sy);
		clickableCurve.setControlX1(c1x);
		clickableCurve.setControlY1(c1y);
		clickableCurve.setControlX2(c2x);
		clickableCurve.setControlY2(c2y);
		clickableCurve.setEndX(ex);
		clickableCurve.setEndY(ey);

		double cps2 = controlPointFeedbackSize / 2;
		firstControlPointHandle.setX(c1x - cps2);
		firstControlPointHandle.setY(c1y - cps2);

		secondControlPointHandle.setX(c2x - cps2);
		secondControlPointHandle.setY(c2y - cps2);

		double bps2 = bendPointFeedbackSize / 2;
		bendPointHandle.setX(ex - bps2);
		bendPointHandle.setY(ey - bps2);

		helperStart.setStartX(sx);
		helperStart.setStartY(sy);
		helperStart.setEndX(c1x);
		helperStart.setEndY(c1y);

		helperEnd.setStartX(c2x);
		helperEnd.setStartY(c2y);
		helperEnd.setEndX(ex);
		helperEnd.setEndY(ey);

		if(arrowWidget != null) {
			arrowWidget.update(ex, ey, c2x, c2y, arrowLength);
		}
		if(label != null) {
			label.setText(labelText);
			double lx = (sx + ex - label.getBoundsInLocal().getWidth()) / 2;
			double ly = (sy + ey) / 2;
			label.setLayoutX(lx);
			label.setLayoutY(ly);
		}
	}

	/** Adds the link nodes of this segment to the link layer node. */
	public void addLinkNodes(DiagramLayers layers, ILinkMVCBundle bundle, String labelText) {
		ILayer linkLayer = layers.getLinkLayer();
		linkLayer.add(visibleCurve, bundle);
		linkLayer.add(clickableCurve, bundle);
		if(arrowWidget != null) {
			linkLayer.add(arrowWidget, bundle);
		}
		if(label != null) {
			label.setText(labelText);
			label.setBackground(
					new Background(new BackgroundFill(WHITE, new CornerRadii(0), new Insets(0))));
			layers.getTextLayer().add(label, bundle);
		}
	}

	/** Removes the link nodes of this segment from the link layer node. */
	public void removeLinkNodes(DiagramLayers layers) {
		ILayer linkLayer = layers.getLinkLayer();
		linkLayer.remove(visibleCurve);
		linkLayer.remove(clickableCurve);
		if(arrowWidget != null) {
			linkLayer.remove(arrowWidget);
		}
		if(label != null) {
			layers.getTextLayer().remove(label);
		}
	}

	/** Adds the feedback nodes of this segment to the feedback layer node. */
	public void addFeedbackNodes(DiagramLayers layers, ILinkMVCBundle bundle) {
		layers.getVisualFeedbackLayer().add(helperStart, bundle);
		layers.getVisualFeedbackLayer().add(helperEnd, bundle);
		layers.getLinkInteractionLayer().add(firstControlPointHandle, bundle);
		layers.getLinkInteractionLayer().add(bendPointHandle, bundle);
		layers.getLinkInteractionLayer().add(secondControlPointHandle, bundle);
	}

	/** Removes the feedback nodes of this segment from the feedback layer node. */
	public void removeFeedbackNodes(DiagramLayers layers) {
		layers.getVisualFeedbackLayer().remove(helperStart);
		layers.getVisualFeedbackLayer().remove(helperEnd);
		layers.getLinkInteractionLayer().remove(firstControlPointHandle);
		layers.getLinkInteractionLayer().remove(bendPointHandle);
		layers.getLinkInteractionLayer().remove(secondControlPointHandle);
	}

	/** Sets all elements to the given mouse transparency. */
	public void setMouseTransparent(boolean transparent) {
		visibleCurve.setMouseTransparent(transparent);
		clickableCurve.setMouseTransparent(transparent);
		bendPointHandle.setMouseTransparent(transparent);
		firstControlPointHandle.setMouseTransparent(transparent);
		secondControlPointHandle.setMouseTransparent(transparent);
		if(arrowWidget != null) {
			arrowWidget.setMouseTransparent(transparent);
		}
		if(label != null) {
			label.setMouseTransparent(transparent);
		}
	}
}
