/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_INCOMING_LINK_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_OUTGOING_LINK_TAG;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base.ContentAnchorageVisualBase;

import javafx.geometry.Insets;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;

/** Base class for {@link IContentAnchorageVisual}s depicted as rectangles (or squares). */
public abstract class RectangularContentAnchorageVisualBase extends ContentAnchorageVisualBase {
	/** Constructor. */
	public RectangularContentAnchorageVisualBase(IContentAnchorageMVCBundle mvcb) {
		super(mvcb, new Rectangle(), new Rectangle());
	}

	/** Returns the visual rectangle. */
	protected final Rectangle getVisualRectangle() {
		// wild cast works: see constructor
		return (Rectangle)getVisualShape();
	}

	/** Returns the hit area rectangle. */
	protected final Rectangle getHitAreaRectangle() {
		// wild cast works: see constructor
		return (Rectangle)getHitAreaShape();
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		if(enableVisual()) {
			Insets i = getInsets();
			Rectangle rectangle = getVisualRectangle();
			rectangle.setX(bounds.getMinX() + i.getLeft());
			rectangle.setY(bounds.getMinY() + i.getTop());
			double w = bounds.getWidth() - i.getLeft() - i.getRight();
			rectangle.setWidth(w);
			double h = bounds.getHeight() - i.getTop() - i.getBottom();
			rectangle.setHeight(h);
			rectangle.setOpacity(getOpacity());
			rectangle.setStrokeWidth(getBorderWidth());
			rectangle.setStrokeType(getBorderType());
			if(getMVCBundle().hasTag(HIGHLIGHT_INCOMING_LINK_TAG)) {
				rectangle.setFill(getHighlightIncomingLinkColor());
				rectangle.setStroke(getHighlightIncomingLinkBorderColor());
			} else if(getMVCBundle().hasTag(HIGHLIGHT_OUTGOING_LINK_TAG)) {
				rectangle.setFill(getHighlightOutgoingLinkColor());
				rectangle.setStroke(getHighlightOutgoingLinkBorderColor());
			} else {
				rectangle.setFill(getFillColor());
				rectangle.setStroke(getBorderColor());
			}
		}

		if(enableInteraction()) {
			// wild cast works: see addNodes(...)
			Rectangle ha = getHitAreaRectangle();
			double hitAreaOutSet = getHitAreaOutset();
			ha.setX(bounds.getMinX() - hitAreaOutSet);
			ha.setY(bounds.getMinY() - hitAreaOutSet);
			ha.setWidth(bounds.getWidth() + 2 * hitAreaOutSet);
			ha.setHeight(bounds.getHeight() + 2 * hitAreaOutSet);
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** Returns the insets for the filled rectangle subtracted from {@link #getDimensions()}. */
	protected Insets getInsets() {
		return new Insets(2);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node n, DiagramCoordinate diagramLocation) {
		if(n == getHitAreaRectangle()) {
			// wild cast works: see addNodes(...)
			Rectangle ha = getHitAreaRectangle();
			double x = diagramLocation.getLocalX(ha);
			double y = diagramLocation.getLocalY(ha);
			double outset = getHitAreaOutset();
			if(x < outset || x > ha.getWidth() - outset || y < outset ||
					y > ha.getHeight() - outset) {
				return EDragGesture.NEW_LINK;
			}
			return EDragGesture.MOVE;
		}
		if(n == getVisualRectangle()) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}
}
