/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.RectangularBorderLocation.getClosestLocationOnBounds;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.visual.rectangular.RectangularContentVisualBase.getRectangularShapeDragGesture;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.layout.IOffsetLayout;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.layout.ISideLayout;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base.ContentVisualBase;

import javafx.collections.ObservableList;
import javafx.geometry.Dimension2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.ClosePath;
import javafx.scene.shape.LineTo;
import javafx.scene.shape.MoveTo;
import javafx.scene.shape.Path;
import javafx.scene.shape.PathElement;

/** Base class for {@link ContentVisualBase content visuals} depicted by a rhomboid. */
public abstract class RhomboidContentVisualBase extends ContentVisualBase {
	/** The default horizontal offset multiplier. */
	private static final double DEFAULT_OFFSET_MULTIPLIER = 3.0;
	/** The default corner arc. */
	private static final Dimension2D DEFAULT_CORNER_ARC = new Dimension2D(15, 15);

	/** Constructor. */
	public RhomboidContentVisualBase(IContentMVCBundle mvcb) {
		super(mvcb, new Path(), new Path());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		double leftX = bounds.getMinX();
		double upperY = bounds.getMinY();
		double width = bounds.getWidth();
		double height = bounds.getHeight();
		if(enableVisual()) {
			Path path = getVisualPathShape();
			path.setOpacity(getOpacity());
			path.setStroke(getBorderColor());
			path.setStrokeWidth(getBorderWidth());
			path.setStrokeType(getBorderType());
			path.setFill(getFillColor());
			makePath(leftX, upperY, width, height, path);
		}

		if(enableInteraction()) {
			Path ha = getHitAreaPathShape();
			double lnk = getHitAreaStartLinkSize();
			double resize = getHitAreaResizeSize();
			double lr = lnk + resize;
			makePath(leftX - lr, upperY - lr, width + 2 * lr, height + 2 * lr, ha);
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** Creates the rhomboid shape. */
	private void makePath(double x, double y, double w, double h, Path p) {
		double offset = getOffset();
		ObservableList<PathElement> pes = p.getElements();
		pes.clear();
		pes.add(new MoveTo(x, y + h));
		pes.add(new LineTo(x + w - offset, y + h));
		pes.add(new LineTo(x + w, y));
		pes.add(new LineTo(x + offset, y));
		pes.add(new ClosePath());
	}

	/** Returns the offset of the upper and lower rhomboid line (0 = rectangle). */
	public double getOffset() {
		return DEFAULT_OFFSET_MULTIPLIER * getViewer().getFeatures().getHorizontalSpacing();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getAnchorageLocation(IContentAnchorageVisual visual) {
		Rectangle2D pb = getCurrentBounds();
		double rhomboidOffset = getOffset();
		Dimension2D anchorageSize = visual.getDimensions();
		Side side = visual.getLayout(ISideLayout.class).getSide();
		double anchorageOffset = visual.getLayout(IOffsetLayout.class).getOffset();
		if(side == Side.LEFT) {
			double lx = rhomboidOffset / 2 - anchorageSize.getWidth() / 2;
			double ly = pb.getHeight() / 2 - anchorageSize.getHeight() / 2;
			return new DiagramCoordinate(lx, ly);
		} else if(side == Side.RIGHT) {
			double lx = pb.getWidth() - rhomboidOffset / 2 - anchorageSize.getWidth() / 2;
			double ly = pb.getHeight() / 2 - anchorageSize.getHeight() / 2;
			return new DiagramCoordinate(lx, ly);
		}
		RectangularBorderLocation rbl =
				new RectangularBorderLocation(side, anchorageOffset, pb, anchorageSize);
		return rbl.getLocation();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		Rectangle2D b = getCurrentBounds();
		Dimension2D dummy = new Dimension2D(1, 1);
		RectangularBorderLocation loc = getClosestLocationOnBounds(
				indication.add(-b.getMinX(), -b.getMinY()), b, dummy, dummy);
		return loc.getLocation().add(b.getMinX(), b.getMinY());
	}

	/** Returns the dimensions of the corner arcs. */
	protected Dimension2D getCornerArcDimensions() {
		return DEFAULT_CORNER_ARC;
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		if(node == getHitAreaShape()) {
			return getRectangularShapeDragGesture(node, diagramLocation, getHitAreaStartLinkSize(),
					getHitAreaResizeSize());
		}
		if(node == getVisualShape() || node == text) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/** Returns the visual path shape. */
	protected final Path getVisualPathShape() {
		// wild cast works: see constructor
		return (Path)getVisualShape();
	}

	/** Returns the hit area path shape. */
	protected final Path getHitAreaPathShape() {
		// wild cast works: see constructor
		return (Path)getHitAreaShape();
	}

	/**
	 * Returns the extension of the hit area in pixel used to trigger link creation. This number
	 * specifies how many pixels beyond the visible border are used to trigger link creation by a
	 * mouse drag gesture.
	 */
	protected double getHitAreaStartLinkSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the size of the hit area in pixel used to resize the content visual. */
	protected double getHitAreaResizeSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getTextAnchorLocation() {
		DiagramCoordinate textAnchorLocation = super.getTextAnchorLocation();
		return textAnchorLocation.add(getOffset() / 2, 0);
	}
}
