/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.widgets;

import static java.lang.Math.atan2;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static javafx.scene.paint.Color.BLACK;

import javafx.collections.ObservableList;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Line;
import javafx.scene.shape.LineTo;
import javafx.scene.shape.MoveTo;
import javafx.scene.shape.Path;
import javafx.scene.shape.PathElement;

/** A widget for displaying an arrow at the end of links. */
public final class LinkArrowWidget extends Group {
	/** The boolean flag for using line or filled triangle arrow head. */
	private final boolean useLineArrows;
	/** The length of the arrow head. */
	private double arrowLength;

	/** Constructor. */
	public LinkArrowWidget(boolean useLineArrows, double anchorX, double anchorY, double directionX,
			double directionY, double arrowLength) {
		this.useLineArrows = useLineArrows;
		this.arrowLength = arrowLength;
		if(useLineArrows) {
			addLines(anchorX, anchorY, directionX, directionY);
		} else {
			addFilledArrow(anchorX, anchorY, directionX, directionY);
		}
	}

	/** Fills the group with a filled arrow widget. */
	private void addFilledArrow(double anchorX, double anchorY, double directionX,
			double directionY) {
		ObservableList<Node> children = getChildren();
		children.clear();
		Path p = new Path();
		p.setFill(BLACK);
		p.setStrokeWidth(1);
		children.add(p);
		updatePath(p, anchorX, anchorY, directionX, directionY);
	}

	/** Updates the {@link Path} primitive for the filled arrow head. */
	private void updatePath(Path p, double anchorX, double anchorY, double directionX,
			double directionY) {
		double angle = atan2(directionX - anchorX, directionY - anchorY);
		double angleLeft = angle - Math.PI / 6;
		double lx = anchorX + sin(angleLeft) * arrowLength;
		double ly = anchorY + cos(angleLeft) * arrowLength;
		double angleRight = angle + Math.PI / 6;
		double rx = anchorX + sin(angleRight) * arrowLength;
		double ry = anchorY + cos(angleRight) * arrowLength;

		ObservableList<PathElement> l = p.getElements();
		l.clear();
		l.add(new MoveTo(anchorX, anchorY));
		l.add(new LineTo(lx, ly));
		l.add(new LineTo(rx, ry));
		l.add(new LineTo(anchorX, anchorY));
	}

	/** Fills the group with two lines forming the arrow head. */
	private void addLines(double anchorX, double anchorY, double directionX, double directionY) {
		ObservableList<Node> children = getChildren();
		children.clear();
		Line leftLine = new Line(0, 0, 0, 0);
		children.add(leftLine);
		Line rightLine = new Line(0, 0, 0, 0);
		children.add(rightLine);
		updateLines(anchorX, anchorY, directionX, directionY);
	}

	/** Updates the locations of the widget's lines */
	private void updateLines(double anchorX, double anchorY, double directionX, double directionY) {
		// get(0) and cast works: see addLines(...)
		Line left = (Line)getChildren().get(0);
		double angle = atan2(directionX - anchorX, directionY - anchorY);
		double angleLeft = angle - Math.PI / 4;
		double lx = anchorX + sin(angleLeft) * arrowLength;
		double ly = anchorY + cos(angleLeft) * arrowLength;
		left.setStartX(anchorX);
		left.setStartY(anchorY);
		left.setEndX(lx);
		left.setEndY(ly);

		// get(1) and cast works: see addLines(...)
		Line right = (Line)getChildren().get(1);
		double angleRight = angle + Math.PI / 4;
		double rx = anchorX + sin(angleRight) * arrowLength;
		double ry = anchorY + cos(angleRight) * arrowLength;
		right.setStartX(anchorX);
		right.setStartY(anchorY);
		right.setEndX(rx);
		right.setEndY(ry);
	}

	/** Updates the arrow head with the new locations. */
	public void update(double anchorX, double anchorY, double directionX, double directionY,
			double arrowLength) {
		this.arrowLength = arrowLength;
		if(useLineArrows) {
			updateLines(anchorX, anchorY, directionX, directionY);
		} else {
			// get(0) and cast works: see AddFilledArrow(...)
			Path p = (Path)getChildren().get(0);
			updatePath(p, anchorX, anchorY, directionX, directionY);
		}
	}

	/** Sets the color of the arrow widget. */
	public void setColor(Paint color) {
		if(useLineArrows) {
			// get(0) and cast works: see addLines(...)
			((Line)getChildren().get(0)).setStroke(color);
			// get(1) and cast works: see addLines(...)
			((Line)getChildren().get(1)).setStroke(color);
		} else {
			// get(0) and cast works: see AddFilledArrow(...)
			Path p = (Path)getChildren().get(0);
			p.setStroke(color);
			p.setFill(color);
		}
	}
}
