/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx.util;

import java.io.IOException;
import java.net.URL;
import java.util.LinkedList;
import java.util.Queue;
import java.util.function.Predicate;

import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;

/**
 * Utility methods for {@link Scene} graph.
 * 
 * @author hoelzl
 * @author diewald
 * @author munaro
 */
public final class JavaFXUtils {

	/** Prevent instantiation. */
	private JavaFXUtils() {
		// Nothing to do
	}

	/**
	 * Loads a JavaFX layout from an {@link FXML} resource and associates it with the given
	 * controller.
	 * 
	 * @param fxmlResource
	 *            {@link URL} of the {@link FXML} resource to load.
	 * @param controller
	 *            Controller to be added to associated with the returned scene.
	 * @return The loaded object hierarchy. See {@link FXMLLoader} for more information.
	 * @throws IOException
	 */
	public static <T> T loadFXML(URL fxmlResource, Object controller) throws IOException {
		FXMLLoader fxmlLoader = new FXMLLoader();
		fxmlLoader.setClassLoader(controller.getClass().getClassLoader());
		fxmlLoader.setController(controller);
		fxmlLoader.setLocation(fxmlResource);

		return fxmlLoader.load();
	}

	/**
	 * Searches the scene graph hierarchy for the nearest parent node of the given class.
	 *
	 * @param start
	 *            the start node
	 * @param clazz
	 *            the target class of the node to be found
	 * @return the nearest parent node
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Node> T findParentNode(Parent start, Class<T> clazz) {
		return (T)findParentNode(start, t -> clazz.isInstance(t));
	}

	/**
	 * Searches the scene graph hierarchy for the nearest child node with the given id.
	 *
	 * @param start
	 *            the start node
	 * @param id
	 *            string identified (field "fx:id" in FXML files).
	 * @return the nearest child node
	 */
	public static Node findParentById(Parent start, String id) {
		return findParentNode(start, t -> t.getId().equals(id));
	}

	/**
	 * Searches the scene graph hierarchy for the nearest parent node with the given id and type.
	 *
	 * @param start
	 *            the start node
	 * @param id
	 *            string identified (field "fx:id" in FXML files).
	 * @param type
	 *            the target class of the node to be found
	 * @return the nearest child node
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Node> T findParentByIdType(Parent start, String id, Class<T> type) {
		return (T)findChildNode(start,
				t -> (type.isAssignableFrom(type)) && (t.getId().equals(id)));
	}

	/**
	 * Searches the scene graph hierarchy for the nearest parent node fulfilling the given
	 * {@link Predicate}.
	 *
	 * @param start
	 *            the start node
	 * @param pred
	 *            predicate evaluated during the parent walk
	 * @return the nearest parent node
	 */
	public static Node findParentNode(Parent start, Predicate<Node> pred) {
		while(start != null) {
			start = start.getParent();
			if(pred.test(start)) {
				return start;
			}
		}
		return null;
	}

	/**
	 * Searches the scene graph hierarchy for the nearest child node of the given class.
	 *
	 * @param start
	 *            the start node
	 * @param clazz
	 *            the target class of the node to be found
	 * @return the nearest child node
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Node> T findChildNode(Parent start, Class<T> clazz) {
		return (T)findChildNode(start, t -> clazz.isInstance(t));
	}

	/**
	 * Searches the scene graph hierarchy for the nearest child node with the given id.
	 *
	 * @param start
	 *            the start node
	 * @param id
	 *            string identified (field "fx:id" in FXML files).
	 * @return the nearest child node
	 */
	public static Node findChildById(Parent start, String id) {
		return findChildNode(start, t -> t.getId().equals(id));
	}

	/**
	 * Searches the scene graph hierarchy for the nearest child node with the given id and type.
	 *
	 * @param start
	 *            the start node
	 * @param id
	 *            string identified (field "fx:id" in FXML files).
	 * @param type
	 *            the target class of the node to be found
	 * @return the nearest child node
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Node> T findChildByIdType(Parent start, String id, Class<T> type) {
		// @CodeFormatterOff
		return (T)findChildNode(start,
				t -> (t.getId() != null
						&& (t.getId().equals(id))
						&& type.isAssignableFrom(t.getClass())));
		// @CodeFormatterOn
	}

	/**
	 * Searches the scene graph hierarchy for the nearest child node of the given fulfilling the
	 * given {@link Predicate}.
	 *
	 * @param start
	 *            the start node
	 * @param pred
	 *            predicate evaluated during the child walk
	 * @return the nearest parent node
	 */
	public static Node findChildNode(Parent start, Predicate<Node> pred) {
		Queue<Node> children = new LinkedList<>();
		children.addAll(start.getChildrenUnmodifiable());
		while(!children.isEmpty()) {
			Node child = children.poll();
			if(pred.test(child)) {
				return child;
			}
			if(child instanceof Parent) {
				children.addAll(((Parent)child).getChildrenUnmodifiable());
			}
		}
		return null;
	}
}
