/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import static org.eclipse.core.resources.ResourcesPlugin.getWorkspace;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.fortiss.tooling.kernel.ToolingKernelActivator;

/**
 * Utility methods used for Eclipse resource storage.
 * 
 * @author hoelzl
 * @author diewald
 */
public final class EMFResourceUtils {
	/**
	 * Stores the given {@link EObject} in the Eclipse workspace creating the given {@link IFile}.
	 * 
	 * @param file
	 *            the target file of path <code>/project/path_to_file</code>.
	 * @param modelRootElement
	 *            the root element of the model to be stored.
	 */
	public static void createNewEclipseWorkspaceProject(IFile file, EObject modelRootElement)
			throws IOException {
		Resource r = new ResourceSetImpl()
				.createResource(URI.createPlatformResourceURI(file.getFullPath().toString(), true));
		r.getContents().add(modelRootElement);
		r.save(buildOptionsMap());
	}

	/**
	 * Loads the root model element from the given {@link URI}. Logs errors to the console and
	 * returns <code>null</code>.
	 * 
	 * @param uri
	 *            the resource {@link URI}.
	 * @return the root {@link EObject}
	 */
	public static EObject loadModelFromFile(URI uri) {
		try {
			return loadModelFromFileWithUnknownFeatures(uri).getFirst();
		} catch(Exception ex) {
			error(ToolingKernelActivator.getDefault(),
					"Failed to load model from " + uri.toString());
		}
		return null;
	}

	/** Refresh the workspace. */
	public static void refreshWorkspace() {
		try {
			// refresh the workspace
			getWorkspace().getRoot().refreshLocal(IResource.DEPTH_INFINITE, null);
		} catch(Exception e) {
			error(ToolingKernelActivator.getDefault(), "Cannot refresh workspace!", e);
		}
	}

	/**
	 * Build the map of options used for storage of EMF models.
	 * 
	 * @return the default option map used by the kernel.
	 */
	public static Map<String, Object> buildOptionsMap() {
		Map<String, Object> options = new HashMap<String, Object>();
		options.put(XMLResource.OPTION_ENCODING, "UTF-8");
		options.put(XMLResource.OPTION_RECORD_UNKNOWN_FEATURE, true);
		options.put(Resource.OPTION_ZIP, false);
		options.put(Resource.OPTION_SAVE_ONLY_IF_CHANGED,
				Resource.OPTION_SAVE_ONLY_IF_CHANGED_MEMORY_BUFFER);

		// Options to increase performance of loading models
		options.put(XMLResource.OPTION_DEFER_ATTACHMENT, true);
		options.put(XMLResource.OPTION_DEFER_IDREF_RESOLUTION, true);

		return options;
	}

	/**
	 * Loads the root model element from the given {@link URI}. Returns {@code null} in case of
	 * errors, and logs errors to the console.
	 * 
	 * @param uri
	 *            the resource {@link URI}.
	 * @return the root {@link EObject}.
	 */
	private static Pair<EObject, Map<EObject, AnyType>>
			loadModelFromFileWithUnknownFeatures(URI uri) {
		return loadModelFromFileWithUnknownFeatures(uri, EObject.class);
	}

	/**
	 * Loads the root model element from the given {@link URI}. Logs errors to the console and
	 * returns <code>null</code>.
	 * 
	 * @param uri
	 *            the resource {@link URI}.
	 * @param topLevelClass
	 *            the type of the top level class to be returned
	 * @return the root {@link EObject} of class type topLevelClass (first root which is found of
	 *         this type is returned)
	 */
	private static Pair<EObject, Map<EObject, AnyType>>
			loadModelFromFileWithUnknownFeatures(URI uri, Class<?> topLevelClass) {
		try {
			ResourceSet rset = new ResourceSetImpl();
			Resource r;
			r = rset.createResource(uri);
			r.load(buildOptionsMap());
			Map<EObject, AnyType> unknownFeatures = r instanceof XMIResource
					? ((XMIResource)r).getEObjectToExtensionMap() : new HashMap<EObject, AnyType>();
			if(!r.getContents().isEmpty()) {
				for(EObject content : r.getContents()) {
					if(topLevelClass.isAssignableFrom(content.getClass())) {
						return new Pair<EObject, Map<EObject, AnyType>>(content, unknownFeatures);
					}
				}
			}
		} catch(IOException ex) {
			error(ToolingKernelActivator.getDefault(),
					"Failed to load model from " + uri.toString(), ex);
		}
		return null;
	}
}
