/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.style;

import static javafx.scene.paint.Color.GREEN;

import java.util.HashMap;
import java.util.Map;

import org.fortiss.tooling.common.ui.javafx.style.FillStyle;
import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.spiderchart.model.AxisBase;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.SpiderChart;

/**
 * Class for all style attributes of the {@link SpiderChart spider chart}.
 * 
 * @author hoelzl
 * @author munaro
 */
public final class ChartStyle {
	/** The map for style information of axes. */
	private Map<AxisBase, AxisStyle> axisStyle = new HashMap<AxisBase, AxisStyle>();
	/** The map for style information of data series. */
	private Map<DataSeries, DataSeriesStyle> dataSeriesStyle =
			new HashMap<DataSeries, DataSeriesStyle>();
	/** Flag for showing the title. */
	private boolean showTitle = true;
	/** The font style for the title. */
	private FontStyle titleStyle;
	/** Flag for showing the legend. */
	private boolean showLegend = true;
	/** The style for the legend. */
	private LegendStyle legendStyle;
	/** The relative chart margin. The default is 10%. */
	private double relativeMargin = 0.1;
	/** The start angle of the first axis (default is 90 degree). */
	private double startAngleDegree = 90.0;
	/**
	 * The flag for using individual segments on each axis. No background spider web will be drawn
	 * if this is <code>true</code>.
	 */
	private boolean useIndividualAxisSegments = false;
	/**
	 * The number of segments to be drawn if {@link #useIndividualAxisSegments} is
	 * <code>false</code>.
	 */
	private int axisSegments = 1;
	/**
	 * The {@link FillStyle} used for the background if {@link #useIndividualAxisSegments} is
	 * <code>false</code>.
	 */
	private FillStyle backgroundFillStyle = new FillStyle(GREEN.brighter());
	/**
	 * The {@link LineStyle} used for the background if {@link #useIndividualAxisSegments} is false.
	 */
	private LineStyle backgroundLineStyle = new LineStyle(GREEN);

	/**
	 * Constructor.
	 */
	public ChartStyle(boolean showTitle, boolean showLegend, boolean useIndividualAxisSegments) {
		this.showTitle = showTitle;
		this.showLegend = showLegend;
		this.useIndividualAxisSegments = useIndividualAxisSegments;
	}

	/** Returns the start angle in degree. */
	public double getStartAngleDegree() {
		return startAngleDegree;
	}

	/** Sets the start angle in degree. */
	public void setStartAngleDegree(double startAngleDegree) {
		this.startAngleDegree = startAngleDegree;
	}

	/**
	 * Sets the relative margin of the chart area. Specify <code>0.1</code> for a margin of 10
	 * percent.
	 */
	public void setRelativeMargin(double relativeMargin) {
		this.relativeMargin = relativeMargin;
	}

	/** Returns the relative margin. */
	public double getRelativeMargin() {
		return relativeMargin;
	}

	/** Returns whether to show the title. */
	public boolean isShowTitle() {
		return showTitle;
	}

	/** Sets the showTitle flag. */
	public void setShowTitle(boolean showTitle) {
		this.showTitle = showTitle;
	}

	/** Returns the title font style. */
	public FontStyle getTitleStyle() {
		return titleStyle;
	}

	/** Sets the title font style. */
	public void setTitleStyle(FontStyle titleStyle) {
		this.titleStyle = titleStyle;
	}

	/** Returns whether to show the legend. */
	public boolean isShowLegend() {
		return showLegend;
	}

	/** Sets showLegend flag. */
	public void setShowLegend(boolean showLegend) {
		this.showLegend = showLegend;
	}

	/** Returns the legend style. */
	public LegendStyle getLegendStyle() {
		return legendStyle;
	}

	/** Sets the legend style. */
	public void setLegendStyle(LegendStyle legendStyle) {
		this.legendStyle = legendStyle;
	}

	/** Returns the style for the given axis. */
	public AxisStyle getAxisStyle(AxisBase axis) {
		return axisStyle.get(axis);
	}

	/** Sets the style for the given axis. */
	public void setAxisStyle(AxisBase axis, AxisStyle style) {
		axisStyle.put(axis, style);
	}

	/** Returns the style for the given data series. */
	public DataSeriesStyle getDataSeriesStyle(DataSeries dataSeries) {
		return dataSeriesStyle.get(dataSeries);
	}

	/** Sets the style for the given data series. */
	public void setDataSeriesStyle(DataSeries dataSeries, DataSeriesStyle style) {
		dataSeriesStyle.put(dataSeries, style);
	}

	/** Returns whether to use individual axis segments. */
	public boolean isUseIndividualAxisSegments() {
		return useIndividualAxisSegments;
	}

	/** Sets the useIndividualAxisSegments flag. */
	public void setUseIndividualAxisSegments(boolean useIndividualAxisSegments) {
		this.useIndividualAxisSegments = useIndividualAxisSegments;
	}

	/** Returns the axis segments. */
	public int getAxisSegments() {
		return axisSegments;
	}

	/** Sets axis segments. */
	public void setAxisSegments(int axisSegments) {
		this.axisSegments = Math.max(1, axisSegments);
	}

	/** Returns the background fill style. */
	public FillStyle getBackgroundFillStyle() {
		return backgroundFillStyle;
	}

	/** Sets the background fill style. */
	public void setBackgroundFillStyle(FillStyle backgroundFillStyle) {
		this.backgroundFillStyle = backgroundFillStyle;
	}

	/** Returns the background line style. */
	public LineStyle getBackgroundLineStyle() {
		return backgroundLineStyle;
	}

	/** Sets the background line style. */
	public void setBackgroundLineStyle(LineStyle backgroundLineStyle) {
		this.backgroundLineStyle = backgroundLineStyle;
	}
}
