/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.jface;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.swt.graphics.Image;

import org.conqat.lib.commons.assertion.CCSMAssert;
import org.conqat.lib.commons.assertion.PreconditionException;
import org.conqat.ide.commons.ui.CCSMCommonsActivator;
import org.conqat.ide.commons.ui.logging.LoggingUtils;

/**
 * Column provider for text columns that uses reflections to determine column
 * value.
 * 
 * @author schwitze
 * @author Florian Deissenboeck
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: B13CE107F8FF5E18A66FFD6C605D0408
 */
public class ReflectiveColumnLabelProvider<T> extends ColumnLabelProvider {

	/** Method used to determine value. */
	private final Method method;

	/** Image returned by {@link #getImage(Object)}. */
	private final Image image;

	/**
	 * Creates label provider. This throws a {@link IllegalArgumentException} if
	 * property was not found for the specified class.
	 * 
	 * @param type
	 *            type this provider is for.
	 * @param methodName
	 *            name of the method to retrieve label from.
	 */
	public ReflectiveColumnLabelProvider(Class<T> type, String methodName) {
		this(type, methodName, null);
	}

	/**
	 * Creates label provider with image. This throws a
	 * {@link PreconditionException} if property was not found for the specified
	 * class.
	 * 
	 * @param type
	 *            type this provider is for.
	 * @param methodName
	 *            name of the method to retrieve label from.
	 * @param imageDescriptor
	 *            ImageDescriptor of the image to display for the column
	 *            optionally. If null then no image is displayed.
	 */
	public ReflectiveColumnLabelProvider(Class<T> type, String methodName,
			ImageDescriptor imageDescriptor) {
		if (imageDescriptor == null) {
			this.image = null;
		} else {
			this.image = imageDescriptor.createImage();
		}
		try {
			method = type.getMethod(methodName);
			if (method.getReturnType() == void.class) {
				throw new PreconditionException("Method has return type void.");
			}
		} catch (SecurityException e) {
			throw new PreconditionException(e.getMessage());
		} catch (NoSuchMethodException e) {
			throw new PreconditionException("Unknown method " + e.getMessage());
		}
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Returns the image created by the {@link ImageDescriptor} given as
	 * constructor argument.
	 */
	@Override
	public Image getImage(Object element) {
		return image;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Disposes the image.
	 */
	@Override
	public void dispose() {
		if (image != null) {
			image.dispose();
		}
		super.dispose();
	}

	/**
	 * Get text via reflection. Throws an {@link AssertionError} if the required
	 * type does not match the type specified for this provider, getter could
	 * not be called or the getter threw an exception.
	 */
	@Override
	public final String getText(Object element) {
		CCSMAssert.isTrue(method.getDeclaringClass().isInstance(element),
				"Expected element to be a " + method.getDeclaringClass());
		try {
			Object result = method.invoke(element);
			if (result == null) {
				return null;
			}
			return result.toString();
		} catch (IllegalArgumentException e) {
			LoggingUtils.error(CCSMCommonsActivator.getDefault(),
					"Error calling method " + method  + ':' + e.getMessage(), e);
		} catch (IllegalAccessException e) {
			LoggingUtils.error(CCSMCommonsActivator.getDefault(),
					"Error calling method " + method  + ':' + e.getMessage(), e);
		} catch (InvocationTargetException e) {
			LoggingUtils.error(CCSMCommonsActivator.getDefault(),
					"Error calling method " + method  + ':' + e.getMessage(), e);
		}
		return "Label provider error. See log for details.";
	}

}